# -*- coding: UTF-8 -*-

import glob
import os
import shutil
import json
import tempfile
import time
import threading


class FileQueueManager:
    """ Класс менеджера очередей на файловой системе """
    def __init__(self, queueRoot=tempfile.gettempdir(), chunkSize=25000, log=None, location='cash'):
        self.queueRoot = queueRoot
        self.chunkSize = chunkSize
        self.log = log
        self.source = ''
        self.isCash = location == 'cash'

        # создание корневой директории очередей
        # здесь можно свалиться, но это нормально
        if not os.path.exists(self.queueRoot):
            os.makedirs(self.queueRoot)

    def getShopDests(self, prefixQueueName):
        """ Получение списка наименований магазинов на основании имён очередей
        Args:
            prefixQueueName(str): префикс очереди
        Returns:
            list: список наименований магазинов
        """
        idirs = glob.iglob(os.path.join(self.queueRoot, '%s*' % prefixQueueName))
        return list(map(lambda x: os.path.basename(x)[len(prefixQueueName):], idirs))

    def removeUncompletedTransactions(self, queueName, source):  # HERE!!! -> долгий метод
        """ Удаление неполных транзакций. Транзакция считается неполной, если у неё transactionId > 0
            и отсутствует сообщение о завершении транзакции.
            Этот метод должен вызывать тот модуль, который создавал транзакцию, указав соответствующий source
        Args:
            queueName(str): имя очереди
            source(str): создатель транзакции(sync-core2, maint, artix-pos, exchangers-upload)
        """
        # 1. получение всех транзакций
        transactions = self._getAllTransactions(self._getRootDir(), queueName)

        # 2. получение множества незавершенных транзакций
        if self.isCash:
            uncompleted = filter(lambda x: not self._isTransactionComplete(queueName, [source], x), transactions)
        else:
            uncompleted = filter(lambda x: not self._isTmpTransactionComplete(queueName, [source], x), transactions)

        # 3. удаление незавершенных транзакций
        for transactionId in uncompleted:
            self.removeTransaction(queueName, transactionId)

        self.source = source

    def beginTransaction(self, queueName):
        """ Начало транзакции. Возвращает идентификатор транзакции
        Args:
            queueName(str): имя очереди
        Returns:
            int: идентификатор транзакции
        """
        transactionId = self._getNextTransactionId(os.path.join(self.queueRoot, queueName))
        if self.log:
            self.log.debug("Инициализация транзакции %s" % transactionId)
        self._beginTransaction(queueName, transactionId)
        return transactionId

    def endTransaction(self, queueName, transactionId, message):
        """ Конец транзакции. Создание сообщения с метаданными транзакции
        Args:
            queueName(str): имя очереди
            transactionId(int): идентификатор транзакции
            message(json): данные сообщения
        """
        if not self.isCash:
            message['properties']['threadId'] = threading.current_thread().ident
        queuePath = os.path.join(self._getRootDir(), queueName)
        if not os.path.exists(queuePath):
            os.makedirs(queuePath)
        with open(os.path.join(queuePath, self._getNextFileName(queuePath, transactionId)), 'w') as trFile:
            json.dump(message, trFile)
        if not self.isCash:
            transactionFiles = self._getAllTransactionFiles(self._getRootDir(), queueName, transactionId)
            for trFile in transactionFiles:
                try:
                    if not os.path.exists(os.path.join(self.queueRoot, queueName)):
                        os.makedirs(os.path.join(self.queueRoot, queueName))
                    shutil.move(trFile, os.path.join(self.queueRoot, queueName, os.path.basename(trFile)))
                except Exception as e:
                    if self.log:
                        self.log.warn(e)
                        self.log.warn('Ошибка перемещения завершенной транзакции %s' % trFile)

    def enqueue(self, queueName, transactionId, message):
        """ Помещение сообщения в очередь. При этом сообщение будет в отдельном файле
        Args:
            queueName(str): имя очереди
            transactionId(int): номер транзакции
            message(dict): сообщение
        """
        queuePath = os.path.join(self._getRootDir(), queueName)
        if not os.path.exists(queuePath):
            os.makedirs(queuePath)
        nextFileName = self._getNextFileName(queuePath, transactionId)
        fileName = os.path.join(queuePath, nextFileName)
        with open(fileName, 'w') as outfile:
            json.dump(message, outfile)

    def getCompleteTransactionId(self, queueName):
        """ Получение идентификатора первой полной тразнакции
        Args:
            queueName(str): имя очереди
        Returns:
            int: идентификатор транзакции или None, если транзакций нет
        """
        allTransactions = sorted(self._getAllTransactions(self.queueRoot, queueName))
        completedTransactions = sorted(list(filter(lambda x: self._isTransactionComplete(queueName,
                ['sync-core2', 'maint', 'artix-pos', 'exchangers-upload'], x), allTransactions)))
        if completedTransactions:
            return completedTransactions[0]
        return None

    def getNextMessage(self, queueName, transactionId):
        """ Получение следующего сообщения транзакции
        Args:
            queueName(str): имя очереди
            transactionId(int): идентификатор транзакции
        Returns:
            dict: следующее сообщение транзакции
        """
        transactionFiles = self._getAllTransactionFiles(self.queueRoot, queueName, transactionId)
        for messagePath in sorted(transactionFiles, key=lambda x: int(os.path.splitext(os.path.basename(x))[0])):
            try:
                with open(messagePath, 'r') as messageFile:
                    yield json.load(messageFile)
            except Exception as e:
                if self.log:
                    self.log.warn('Ошибка чтения файла %s. %s' % (messagePath, str(e)))

    def getTransactionProperties(self, queueName, transactionId):
        if self.log:
            self.log.info('Получение свойств транзакции %s из очереди %s' % (transactionId, queueName))
        transactionFiles = self._getAllTransactionFiles(self.queueRoot, queueName, transactionId)
        if not transactionFiles:
            return None
        lastFileName = max(transactionFiles, key=lambda x: int(os.path.splitext(os.path.basename(x))[0]))
        with open(lastFileName, 'r') as lastFile:
            data = json.load(lastFile)
            return data.get('properties')

    def dequeue(self, messagePath):
        """ Удаление сообщения по его полному пути
        Args:
            messagePath(str): полный путь до сообщения
        """
        if self.log:
            self.log.debug('Удаление сообщения "%s"' % messagePath)
        if messagePath and os.path.exists(messagePath):
            os.remove(messagePath)

    def removeTransaction(self, queueName, transactionId):
        """ Удаление всех файлов транзакции по идентификатору транзакции
        Args:
            queueName(str): имя очереди
            transactionId(int): идентификатор транзакции
        """
        self._removeTransaction(self.queueRoot, queueName, transactionId)
        if not self.isCash:
            self._removeTransaction(self._getRootDir(), queueName, transactionId)
            
    def _getAllTransactions(self, queueRoot, queueName):
        files = glob.glob(os.path.join(queueRoot, queueName, '*.objs'))
        return set(map(lambda x: int(os.path.splitext(os.path.basename(x))[0]) >> self._getOffsetTransaction(), files))
    
    def _getAllTransactionFiles(self, queueRoot, queueName, transactionId):
        files = glob.glob(os.path.join(queueRoot, queueName, '*.objs'))
        return filter(lambda x: (int(os.path.splitext(os.path.basename(x))[0]) >> self._getOffsetTransaction()) == transactionId, files)
                    
    def _removeTransaction(self, queueRoot, queueName, transactionId):
        if self.log:
            self.log.info('Удаление транзакции с идентификатором %s' % transactionId)
        transactionFiles = self._getAllTransactionFiles(queueRoot, queueName, transactionId)
        for trFile in transactionFiles:
            try:
                os.remove(trFile)
            except:
                if self.log:
                    self.log.warn('Ошибка удаления файла %s' % trFile)

    def _getNextFileName(self, queuePath, transactionId):
        """ Получение имени следующего файла данных
        Args:
            queuePath(str): полный путь до очереди
            transactionId(int): номер транзакции
        Returns:
            str: имя следующего файла данных
        """
        files = glob.glob(os.path.join(queuePath, '*.objs'))
        transactionFiles = filter(lambda x: (int(os.path.splitext(os.path.basename(x))[0]) >> self._getOffsetTransaction()) == transactionId, files)
        nextNumber = max(list(map(lambda x: int(os.path.splitext(os.path.basename(x))[0]), transactionFiles)) + [transactionId << self._getOffsetTransaction()]) + 1
        return '%d.objs' % nextNumber

    def _isTransactionComplete(self, queueName, source, transactionId):
        """ Проверка того, что транзакция является полной
        Args:
            queueName(str): имя очереди
            source(list): список создателей транзакции(sync-core2, maint, artix-pos, exchangers-upload)
            transactionId(int): идентификатор транзакции
        Returns:
            bool: True - транзакция полна, False - иначе
        """
        transactionFiles = self._getAllTransactionFiles(self.queueRoot, queueName, transactionId)
        if not transactionFiles:
            return True
        maxFile = max(transactionFiles, key=lambda x: int(os.path.splitext(os.path.basename(x))[0]))
        if not os.path.exists(maxFile):
            return False
        with open(maxFile, 'r') as mf:
            try:
                data = json.load(mf)
                if data['properties']['source'] not in source or data['properties']['control'] == 'end_tran':
                    return True
            except Exception:
                pass
        return False
    
    def _isTmpTransactionComplete(self, queueName, source, transactionId):
        """ Проверка того, что транзакция является полной и принадлежит текущему потоку и сервису
        Args:
            queueName(str): имя очереди
            source(list): список создателей транзакции(sync-core2, maint, artix-pos, exchangers-upload)
            transactionId(int): идентификатор транзакции
        Returns:
            bool: True - транзакция полна, False - иначе
        """
        transactionFiles = self._getAllTransactionFiles(self._getRootDir(), queueName, transactionId)
        if not transactionFiles:
            return True
        maxFile = max(transactionFiles, key=lambda x: int(os.path.splitext(os.path.basename(x))[0]))
        if not os.path.exists(maxFile):
            return False
        with open(maxFile, 'r') as mf:
            try:
                data = json.load(mf)
                if data['properties']['source'] not in source or data['properties']['control'] == 'end_tran' or data['properties']['threadId'] != threading.current_thread().ident:
                    return True
            except Exception:
                """
                Ошибка при чтении говорит что транзакция в этот момент либо еще пишется другим потоком, 
                либо это вообще не транзакция, в любом случаи эти файлы нас не интересуют
                """
                return True
        return False

    def _beginTransaction(self, queueName, transactionId):
        """ Инициализация транзакции. Создание сообщения с метаданными транзакции
        Args:
            queueName(str): имя очереди
            transactionId(int): идентификатор транзакции
        """
        prop = {'type': 'control', 'control': 'begin_tran', 'source': self.source, 'transactionId': transactionId}
        if not self.isCash:
            prop['threadId'] = threading.current_thread().ident
        
        queuePath = os.path.join(self._getRootDir(), queueName)
        if not os.path.exists(queuePath):
            os.makedirs(queuePath)
        with open(os.path.join(queuePath, self._getNextFileName(queuePath, transactionId)), 'w') as trFile:
            json.dump({'properties': prop}, trFile)

    def _getNextTransactionId(self, queuePath):
        """ Получение номера следующей транзакции
        Args:
            queuePath(str): полный путь до очереди
        Returns:
            int: номер следующей транзакции
        """
        if self.isCash:
            files = glob.glob(os.path.join(queuePath, '*.objs'))
            nextNumber = (max(list(map(lambda x: int(os.path.splitext(os.path.basename(x))[0]), files)) + [0]) >> self._getOffsetTransaction()) + 1
            return nextNumber
        else:
            return int(time.time() * 10000)

    def _getRootDir(self):
        """
        Возвращает корневую директорию для очереди
        Returns:
            str: корневая директория
        """
        return self.queueRoot if self.isCash else tempfile.gettempdir() + self.queueRoot

    def _getOffsetTransaction(self):
        """
        Получить смещение для идентфиикатора транзации
        Returns:
            int: смещение
        """
        return 32 if self.isCash else 48
